"""
    System exception
"""


class VaildateUIDError(Exception):
    """ Raised when the UID is not valid. """


"""
    EnkaNetwork network exception
"""


class EnkaNetworError(Exception):
    """Base class for EnkaNetwork errors."""


class NetworkError(EnkaNetworError):
    """Base class for exceptions due to networking errors."""


class TimedOut(NetworkError):
    """Raised when a request took too long to finish."""


class EnkaServerError(EnkaNetworError):
    """ Exception that's raised for when status code 500 occurs."""


class EnkaServerMaintanance(EnkaNetworError):
    """ Exception that's raised when status code 424 occurs. """


class EnkaServerRateLimit(EnkaNetworError):
    """ Exception that's raised when status code 429 occurs."""


class EnkaServerUnknown(EnkaNetworError):
    """ Exception that's raised when status code 503 occurs. """


"""
    Github
"""


class HTTPException(Exception):
    """Base class for EnkaNetwork errors."""


"""
    EnkaNetwork response error
"""


class ProfileNotFounded(Exception):
    """ Raised when the profile is not found. """


class BuildNotPublicData(Exception):
    """ Raised when the profile hoyos has public to hidden """


class EnkaValidateFailed(EnkaNetworError):
    """ Exception that's raised for when status code 400 occurs."""


class EnkaPlayerNotFound(Exception):
    """ Raised when the UID is not found. """


ERROR_ENKA = {
    400: [VaildateUIDError, "Validate UID {uid} failed."],
    404: [EnkaPlayerNotFound, "Player ID {uid} not found. Please check your UID / Username"],  # noqa
    429: [EnkaServerRateLimit, "Enka.network has been rate limit this path"],
    424: [EnkaServerMaintanance, "Enka.Network doing maintenance server. Please wait took 5-8 hours or 1 day"],  # noqa
    500: [EnkaServerError, "Enka.network server has down or Genshin server broken."],  # noqa
    503: [EnkaServerUnknown, "I screwed up massively"]
}



import logging

from pydantic import BaseModel, Field
from typing import Any, List, Union, Optional

from .utils import IconAsset

from ..enum import EquipmentsType, DigitType, EquipType
from ..assets import Assets

LOGGER = logging.getLogger(__name__)

__all__ = (
    'EquipmentsStats',
    'EquipmentsDetail',
    'Equipments'
)


class EquipmentsStats(BaseModel):
    prop_id: str = ""
    type: DigitType = DigitType.NUMBER
    name: str = ""
    value: Union[int, float] = Field(0, alias="statValue")

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        LOGGER.debug("=== Fight prop ===")

        if "mainPropId" in data:
            self.prop_id = str(data["mainPropId"])
            fight_prop = Assets.get_hash_map(str(data["mainPropId"]))
        else:
            self.prop_id = str(data["appendPropId"])
            fight_prop = Assets.get_hash_map(str(data["appendPropId"]))

        prod_id = ["HURT", "CRITICAL", "EFFICIENCY", "PERCENT", "ADD"]

        if self.prop_id.split("_")[-1] in prod_id:
            self.value = float(data["statValue"])
            self.type = DigitType.PERCENT

        if not fight_prop:
            return

        self.name = fight_prop


class EquipmentsDetail(BaseModel):
    """
        Custom data
    """
    name: str = ""  # Get from name hash map
    artifact_name_set: str = ""  # Name set artifacts
    artifact_type: EquipType = EquipType.Unknown  # Type of artifact
    icon: IconAsset = None
    rarity: int = Field(0, alias="rankLevel")
    mainstats: EquipmentsStats = Field(None, alias="reliquaryMainstat")
    substats: List[EquipmentsStats] = []

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        if data["itemType"] == "ITEM_RELIQUARY":  # AKA. Artifact
            LOGGER.debug("=== Artifact ===")
            self.artifact_type = EquipType(data["equipType"])
            # Sub Stats
            for stats in data["reliquarySubstats"] if "reliquarySubstats" in data else []:
                self.substats.append(EquipmentsStats.parse_obj(stats))

        if data["itemType"] == "ITEM_WEAPON":  # AKA. Weapon
            LOGGER.debug("=== Weapon ===")

            # Main and Sub Stats
            self.mainstats = EquipmentsStats.parse_obj(
                data["weaponStats"][0])
            for stats in data["weaponStats"][1:]:
                self.substats.append(EquipmentsStats.parse_obj(stats))

        _name = Assets.get_hash_map(data.get("nameTextMapHash"))
        if "setNameTextMapHash" in data:
            _artifact_name_set = Assets.get_hash_map(str(data["setNameTextMapHash"]))
            self.artifact_name_set = _artifact_name_set or ""

        self.name = _name if _name is not None else ""

    class Config:
        use_enum_values = True

class EquipmentsProps(BaseModel):
    id: int = 0
    prop_id: str = ''
    name: str = ''
    digit: DigitType = DigitType.NUMBER
    value: Union[int, float] = 0
    
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.name = Assets.get_hash_map(self.prop_id)

    def get_value_symbol(self):
        return f"{self.value}{'%' if self.digit == DigitType.PERCENT else ''}"

    def get_full_name(self):
        raw = self.value
        return f"{self.name}{str(raw) if raw < 0 else '+'+str(raw)}{'%' if self.digit == DigitType.PERCENT else ''}"

class Equipments(BaseModel):
    """
        API Response data
    """
    id: int = Field(0, alias="itemId")
    detail: EquipmentsDetail = Field({}, alias="flat")

    """
        Custom data
    """
    level: int = 0  # Get form key "reliquary" and "weapon"
    max_level: int = 0
    # Type of equipments (Ex. Artifact, Weapon)
    type: EquipmentsType = EquipmentsType.UNKNOWN
    refinement: int = 1  # Refinement  of equipments (Weapon only)
    ascension: int = 0  # Ascension (Weapon only)
    props: List[EquipmentsProps] = []

    class Config:
        use_enum_values = True

    def __init__(self, **data: Any) -> None:
        data["flat"]["icon"] = IconAsset(filename=data["flat"]["icon"])
        super().__init__(**data)

        if data["flat"]["itemType"] == "ITEM_RELIQUARY":  # AKA. Artifact
            self.type = EquipmentsType.ARTIFACT
            self.level = data["reliquary"]["level"] - 1
            self.max_level = 4 * data["flat"]["rankLevel"]

            for props in data["reliquary"].get("appendPropIdList", []):
                props_info = Assets.artifact_props(props)
                if props_info: 
                    self.props.append(EquipmentsProps(**{
                        "id": props_info.id,
                        "prop_id": props_info.type,
                        "digit": DigitType.PERCENT if props_info.digit == 'PERCENT' else DigitType.NUMBER,
                        "value": props_info.value
                    }))

        if data["flat"]["itemType"] == "ITEM_WEAPON":  # AKA. Weapon
            self.type = EquipmentsType.WEAPON
            self.level = data["weapon"]["level"]

            if "affixMap" in data["weapon"]:
                self.refinement = data["weapon"]["affixMap"][
                                                   list(data["weapon"]["affixMap"].keys())[0]] + 1

            if "promoteLevel" in data["weapon"]:
                self.ascension = data["weapon"]["promoteLevel"]
                self.max_level = (self.ascension * 10) + (
                    10 if self.ascension > 0 else 0) + 20
                
                if self.ascension >= 2:
                    self.detail.icon = IconAsset(filename=self.detail.icon.filename + "_Awaken")
            


from __future__ import annotations

from typing import Any, Dict, List, TypedDict, Union

__all__ = ('EnkaNetwork', 'Default')

class profile_picture(TypedDict):
    avatarId: int

class avatar_info(profile_picture):
    level: int

class player_info(TypedDict):
    nickname: str
    level: int
    signature: str
    worldLevel: int
    nameCardId: int
    finishAchievementNum: int
    towerFloorIndex: int
    towerLevelIndex: int
    showAvatarInfoList: List[avatar_info]
    showNameCardIdList: List[int]
    profilePicture: profile_picture

class avatar_info_full(TypedDict):
    avatarId: int
    propMap: Dict[Dict, Dict[str, Any]]
    talentIdList: List[int]
    fightPropMap: Dict[str, Union[int, float]]
    skillDepotId: int
    inherentProudSkillList: List[int]
    skillLevelMap: Dict[str, int]
    equipList: List[Dict[str, Any]]  # TODO: check if this is correct
    fetterInfo: Dict[str, int]

class EnkaNetworkResponse(TypedDict):
    playerInfo: player_info
    avatarInfoList: avatar_info_full
    ttl: int

class EnkaNetwork(TypedDict):
    status: int
    content: EnkaNetworkResponse

class Default:
    status: int
    content: Any



from pydantic import BaseModel
from typing import Any

from ..utils import create_ui_path

class IconAsset(BaseModel):
    filename: str = ""
    url: str = ""
    
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.url = create_ui_path(self.filename)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2022-present M-307
Copyright (c) 2024-present DeviantUa

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE."""

from __future__ import annotations

import asyncio
import aiohttp
import logging
import warnings

from typing import (
    Any,
    Optional,
    TypeVar,
    Coroutine,
    Dict,
    Union,
    TYPE_CHECKING
)
from . import utils
from .config import Config
from .utils import MISSING, RETRY_MAX
from .exception import (
    VaildateUIDError,
    HTTPException,
    EnkaServerError,
    EnkaServerUnknown,
    TimedOut,
    ERROR_ENKA
)

if TYPE_CHECKING:
    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]
    from .types.enkanetwork import (
        EnkaNetwork as EnkaNetworkPayload,
        Default as DefaultPayload,
    )


class Route:
    def __init__(
        self,
        method: str,
        path: str,
        endpoint: str = 'enka',
        username: Optional[str] = None
    ) -> None:
        self.method = method
        self.url = ''
        self.username = username

        if endpoint == 'enka':
            self.url: str = Config.ENKA_PROTOCOL + "://" + Config.ENKA_URL + path
        else:
            self.url: str = Config.ASSETS_PROTOCOL + "://" + Config.ASSETS_URL + path


class HTTPClient:
    LOGGER = logging.getLogger(__name__)

    def __init__(self, *, key: str = '', agent: str = '', timeout: int = 5) -> None:  # noqa
        self.__session: aiohttp.ClientSession = MISSING
        self.__headers: Dict = {}
        self.__timeout = timeout or 10

        # Init User Agent
        if agent != '':
            Config.init_user_agent(agent)

        if key != '':
            warnings.warn("'key' has depercated.")

    async def close(self) -> None:
        if self.__session is not MISSING:
            await self.__session.close()
            self.__session = MISSING
            self.LOGGER.debug('Session closed')
        else:
            self.LOGGER.debug('Session already closed')

    async def request(self, route: Route, **kwargs: Any) -> Any:
        method = route.method
        url = route.url
        username = route.username

        self.__headers.clear()
        if Config.USER_AGENT != '':
            self.__headers['User-Agent'] = Config.USER_AGENT

        kwargs['headers'] = self.__headers

        response: Optional[aiohttp.ClientResponse] = None
        data: Optional[Union[Dict[str, Any], str]] = None

        if self.__session is MISSING:
            self.__session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=self.__timeout))

        for tries in range(RETRY_MAX):
            try:
                async with self.__session.request(method, url, **kwargs) as response: # noqa

                    _host = response.host
                    if 300 > response.status >= 200:
                        data = await utils.to_data(response)

                        self.LOGGER.debug(
                            '%s %s has received %s', method, url, data)
                        return data

                    if _host == Config.ENKA_URL:
                        err = ERROR_ENKA.get(response.status, None)
                        if err:
                            raise err[0](err[1].format(uid=username))

                        raise EnkaServerUnknown(
                            f"Unknow error HTTP status: {response.status}")

                    if response.status >= 400:
                        self.LOGGER.warning(
                            f"Failure to fetch {url} ({response.status}) Retry {tries} / {RETRY_MAX}") # noqa
                        if tries > RETRY_MAX:
                            raise HTTPException(f"Failed to download {url}")
                        await asyncio.sleep(1)  # 1 + tries * 2
                        continue

                    raise HTTPException("Unknown error")

            except OSError as e:
                # Connection reset by peer
                if tries < 4 and e.errno in (54, 10054):
                    await asyncio.sleep(1 + tries * 2)
                    continue
                raise TimedOut("Timeout from enka.network")

        if response is not None:
            # We've run out of retries, raise.
            if response.status >= 500:
                raise EnkaServerError("Server error")

            raise HTTPException("Unknown error")

        raise RuntimeError('Unreachable code in HTTP handling')

    def fetch_user_by_uid(
        self,
        uid: Union[str, int],
        *,
        info: bool = False
    ) -> Response[EnkaNetworkPayload]:
        if not utils.validate_uid(str(uid)):
            raise VaildateUIDError(
                "Validate UID failed. Please check your UID.")

        r = Route(
            'GET',
            f'/api/uid/{uid}' + ("?info" if info else ""),
            endpoint='enka',
            username=uid
        )
        return self.request(r)

    def fetch_user_by_username(
        self,
        username: Union[str, int]
    ) -> Response[EnkaNetworkPayload]:
        r = Route(
            'GET',
            f'/api/profile/{username}',
            endpoint='enka',
            username=username
        )
        return self.request(r)

    def fetch_hoyos_by_username(
        self,
        username: Union[str, int],
        metaname: str = "",
        show_build: bool = False
    ):
        r = Route(
            'GET',
            f'/api/profile/{username}/hoyos'
            + (f"/{metaname}" if metaname != '' else '')
            + ('/builds' if (show_build and metaname != '') else ''),
            endpoint='enka',
            username=username
        )
        return self.request(r)

    def fetch_asset(self, folder: str, filename: str) -> Response[DefaultPayload]: # noqa
        r = Route(
            'GET',
            f'/mrwan200/enkanetwork.py-data/master/exports/{folder}/{filename}', # noqa
            endpoint='assets'
        )
        return self.request(r)

    async def read_from_url(self, url: str) -> bytes:
        async with self.__session.get(url) as resp:
            if resp.status == 200:
                return await resp.read()
            elif resp.status == 404:
                raise HTTPException(resp, 'asset not found')
            else:
                raise HTTPException(resp, 'failed to get asset')



from pydantic import BaseModel, Field
from typing import List, Any, Optional, Union

from ..enum import ElementType
from .utils import IconAsset

__all__ = (
    'NamecardAsset',
    'CharacterIconAsset',
    'CharacterSkillAsset',
    'CharacterConstellationsAsset',
    'CharacterCostume',
    'CharacterAsset'
)


class NamecardAsset(BaseModel):
    """ Namecards (Assets)

    Attributes
    ------------
    id: :class:`int`
        Namecard ID
    hash_id: :class:`str`
        Namecard hash id
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    banner: :class:`IconAsset`
        A banner assets. Please refers in `IconAsset` class
    navbar: :class:`IconAsset`
        A navbar assets. Please refers in `IconAsset` class
    """

    id: int = 0
    hash_id: Union[str, int] = Field("", alias="nameTextMapHash")
    icon: IconAsset
    banner: IconAsset
    navbar: IconAsset

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        
        self.hash_id = str(self.hash_id)

class CharacterIconAsset(BaseModel):
    """ Character Icon (Assets)

    Attributes
    ------------
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    side: :class:`IconAsset`
        A navbar assets. Please refers in `IconAsset` class
    banner: :class:`IconAsset`
        A banner assets. Please refers in `IconAsset` class
    card: :class:`IconAsset`
        A navbar assets. Please refers in `IconAsset` class
    """
    icon: IconAsset
    side: IconAsset
    banner: IconAsset
    card: IconAsset


class CharacterSkillAsset(BaseModel):
    """ Character Skill(s) (Assets)

    Attributes
    ------------
    id: :class:`int`
        Character skill(s) ID
    pround_map: :class:`int`
        pround map for a booest skill by constellation
    hash_id: :class:`str`
        Skill(s) hash id
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    """
    id: int = 0,
    pround_map: int = 0,
    hash_id: Union[str, int] = Field("", alias="nameTextMapHash")
    icon: IconAsset = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        
        self.hash_id = str(self.hash_id)

class CharacterConstellationsAsset(BaseModel):
    """ Character Constellations (Assets)

    Attributes
    ------------
    id: :class:`int`
        Character constellations ID
    hash_id: :class:`str`
        Constellations hash id
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    """
    id: int = 0
    hash_id: Union[str, int] = Field("", alias="nameTextMapHash")
    icon: IconAsset = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        
        self.hash_id = str(self.hash_id)

class CharacterCostume(BaseModel):
    """ Character Costume (Assets)

    Attributes
    ------------
    id: :class:`int`
        Costume ID
    hash_id: :class:`str`
        Costume hash id
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    """
    id: int = 0
    images: CharacterIconAsset = None
    
class CharacterAvatar(BaseModel):
    """ Character Costume (Assets)

    Attributes
    ------------
    id: :class:`int`
        Costume ID
    hash_id: :class:`str`
        Costume hash id
    icon: :class:`IconAsset`
        A icon assets. Please refers in `IconAsset` class
    """
    id: int = 0
    images: CharacterIconAsset = None

class AritfactProps(BaseModel):
    id: int = 0
    type: str = Field('', alias='propType')
    digit: str = Field('DIGIT', alias='propDigit')
    value: Union[int, float] = Field(0, alias='propValue')
    
class CharacterAsset(BaseModel):
    """ Character (Assets)

    Attributes
    ------------
    id: :class:`int`
        Avatar ID
    rarity: :class:`int`
        Character rarity (5 stars or 4stars)
    hash_id: :class:`str`
        Character hash id
    element: :class:`ElementType`
        Character element type
    images: :class:`CharacterIconAsset`
        Character image data. Please refers in `CharacterIconAsset`
    skill_id: :class:`int`
        Character skill ID
    skills: List[:class:`int`]
        Character skill data list
    constellations:  List[:class:`int`]
        Character constellations data list
    """
    id: int = 0
    rarity: int = 0
    hash_id: Union[str, int] = Field("", alias="nameTextMapHash")
    element: ElementType = ElementType.Unknown
    images: CharacterIconAsset = None
    skill_id: int = 0
    skills: List[int] = []
    constellations: List[int] = Field([], alias="talents")

    class Config:
        use_enum_values = True

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self.hash_id = str(self.hash_id)
        self.element = ElementType(
            data["costElemType"]) if data["costElemType"] != "" else ElementType.Unknown
        self.rarity = 5 if data["qualityType"].endswith("_ORANGE") else 4
    


from pydantic import BaseModel

from .players import PlayerInfo
from .build import Builds

class PlayerHoyos(BaseModel):
    uid_public: bool = False
    public: bool = False
    verified: bool = False
    player_info: PlayerInfo
    hash: str
    region: str
    order: str

    builds: Builds = None

    


from __future__ import annotations

import json
import logging
import os

from typing import Union, Optional, Type, TYPE_CHECKING, List, Any, Dict

from .http import HTTPClient
from .model.base import (
    EnkaNetworkResponse,
    EnkaNetworkProfileResponse
)
from .enkanetwork_update import enka_update
from .model.hoyos import PlayerHoyos
from .model.build import Builds

from .utils import update_pfps
from .assets import Assets
from .enum import Language
from .cache import Cache, StaticCache
from .config import Config
from .tools import (
    merge_raw_data
)

if TYPE_CHECKING:
    from typing_extensions import Self
    from types import TracebackType

__all__ = ("EnkaNetworkAPI",)


class EnkaNetworkAPI:
    """

    A library for API wrapper player by UID / Username
    from https://enka.network

    Parameters
    ------------
    lang: :class:`str`
        Init default language
    debug: :class:`bool`
        If set to `True`. In request data or get assets.
        It's will be shown log processing
    key: :class:`str`
        Depercated
    cache: :class:`bool`
        If set to `True`. In response data will be cache data
    user_agent: :class:`str`
        User-Agent for speical to request Enka.Network
    timeout: :class:`int`
        Request timeout to Enka.Network

    Attributes
    ------------
    assets: :class:`Assets`
        Assets character / artifact / namecards / language / etc. data
    http: :class:`HTTPClient`
        HTTP for request and handle data
    lang: :class:`Language`
        A default language

    Example
    ------------
    ```py
    import asyncio

    from enkanetwork import EnkaNetworkAPI

    client = EnkaNetworkAPI(lang="th",user_agent="SpeicalAgent/1.0")

    async def main():
        async with client:
            data = await client.fetch_user(843715177)
            print(data.player.nickname)

    asyncio.run(main())
    ```
    """
    LOGGER = logging.getLogger(__name__)

    def __init__(
        self,
        *,
        lang: str = "en",
        debug: bool = False,
        key: str = "",
        cache: bool = True,
        user_agent: str = "",
        timeout: int = 10
    ) -> None:  # noqa: E501
        # Logging
        logging.basicConfig()
        logging.getLogger("enkanetwork").setLevel(logging.DEBUG if debug else logging.ERROR)  # noqa: E501

        # Set language and load config
        self.assets = Assets(lang)

        # Cache
        self._enable_cache = cache
        if self._enable_cache:
            Config.init_cache(StaticCache(1024, 60 * 1))

        # http client
        self.__http = HTTPClient(key=key, agent=user_agent, timeout=timeout)
        self._closed = False

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> None:
        self._closed = True
        if self._closed:
            await self.__http.close()

    def is_closed(self) -> bool:
        return self._closed

    @property
    def http(self) -> HTTPClient:
        return self.__http

    @http.setter
    def http(self, http: HTTPClient) -> None:
        self.__http = http

    @property
    def lang(self) -> Language:
        return self.assets.LANGS

    @lang.setter
    def lang(self, lang: Language) -> None:
        self.assets._set_language(lang)

    def set_cache(self, cache: Cache) -> None:
        Config.init_cache(cache)

    async def set_language(self, lang: Language) -> None:
        self.lang = lang

    async def fetch_user_by_uid(
        self,
        uid: Union[str, int],
        *,
        info: bool = False
    ) -> EnkaNetworkResponse:
        """ Fetch user profile by UID

        Parameters
        ------------
        uid: Union[:class:`str`,:class:`int`]
            UID player in-game
        info: :class:`bool`
            If set to `True`. It's will response player info only

        Raises
        ------------
        VaildateUIDError
            Player UID empty/format has incorrect.
        EnkaPlayerNotFound
            Player UID doesn't not exists in-game
        EnkaServerRateLimit
            Enka.Network has been rate limit
        EnkaServerMaintanance
            Enka.Network has maintenance server
        EnkaServerError
            Enka.Network has server error (The reason normal is `general`)
        EnkaServerUnknown
            Enka.Network has error another

        Returns
        ------------
        :class:`EnkaNetworkResponse`
            The response player data
        """

        # Loda cache
        cache = await self.__get_cache(uid)
        if cache:
            return EnkaNetworkResponse.parse_obj(cache)

        data = await self.__http.fetch_user_by_uid(uid, info=info)
        data = self.__format_json(data)

        # Return data
        self.LOGGER.debug("Parsing data...")

        # Store cache
        await self.__store_cache(uid, data)

        if "owner" in data:
            data["owner"] = {
                **data["owner"],
                "builds": await self.fetch_builds(
                    profile_id=data["owner"]["username"],
                    metaname=data["owner"]["hash"]
                )
            }
        return EnkaNetworkResponse.parse_obj(data)

    async def fetch_user_by_username(
        self,
        profile_id: Optional[str]
    ) -> EnkaNetworkProfileResponse:
        """ Fetch user profile by Username / patreon ID

        Parameters
        ------------
        profile_id: Optional[:class:`str`]
            Username / patreon ID has subscriptions in Enka.Network

        Raises
        ------------
        EnkaPlayerNotFound
            Player UID doesn't not exists in-game
        EnkaServerRateLimit
            Enka.Network has been rate limit
        EnkaServerMaintanance
            Enka.Network has maintenance server
        EnkaServerError
            Enka.Network has server error (The reason normal is `general`)
        EnkaServerUnknown
            Enka.Network has error another

        Returns
        ------------
        :class:`EnkaNetworkProfileResponse`
            The response profile / hoyos and builds data
        """
        # Loda cache
        cache = await self.__get_cache(profile_id)
        if cache:
            return EnkaNetworkProfileResponse.parse_obj(cache)

        data = await self.__http.fetch_user_by_username(profile_id)
        data = self.__format_json(data)

        self.LOGGER.debug("Parsing data...")

        # Store cache
        await self.__store_cache(profile_id, data)

        # Fetch hoyos and build(s)
        data = {
            **data,
            "hoyos": await self.fetch_hoyos_by_username(profile_id)
        }

        return EnkaNetworkProfileResponse.parse_obj(data)

    async def fetch_hoyos_by_username(
        self,
        profile_id: Optional[str]
    ) -> List[PlayerHoyos]:
        """ Fetch hoyos user data by Username / patreon ID

        Parameters
        ------------
        profile_id: Optional[:class:`str`]
            Username / patreon ID has subscriptions in Enka.Network

        Raises
        ------------
        EnkaPlayerNotFound
            Player UID doesn't not exists in-game
        EnkaServerRateLimit
            Enka.Network has been rate limit
        EnkaServerMaintanance
            Enka.Network has maintenance server
        EnkaServerError
            Enka.Network has server error (The reason normal is `general`)
        EnkaServerUnknown
            Enka.Network has error another

        Returns
        ------------
        List[:class:`PlayerHoyos`]
            A response hoyos player data
        """
        key = profile_id + ":hoyos"

        # Loda cache
        cache = await self.__get_cache(key)
        if cache:
            return self.__format_hoyos(profile_id, cache)

        data = await self.__http.fetch_hoyos_by_username(profile_id)
        data = self.__format_json(data)
        self.LOGGER.debug("Parsing data...")

        # Store cache
        await self.__store_cache(key, data)

        return await self.__format_hoyos(profile_id, data)

    async def fetch_builds(
        self,
        *,
        profile_id: Optional[str],
        metaname: Optional[str]
    ) -> Builds:
        """ Fetch hoyos build(s) data

        Parameters
        ------------
        profile_id: Optional[:class:`str`]
            Username / patreon ID has subscriptions in Enka.Network
        metaname: Optional[:class:`str`]
            Metaname from hoyos data or owner tag in hash field

        Raises
        ------------
        EnkaPlayerNotFound
            Player UID doesn't not exists in-game
        EnkaServerRateLimit
            Enka.Network has been rate limit
        EnkaServerMaintanance
            Enka.Network has maintenance server
        EnkaServerError
            Enka.Network has server error (The reason normal is `general`)
        EnkaServerUnknown
            Enka.Network has error another

        Returns
        ------------
        :class:`Builds`
            A response builds data
        """
        key = profile_id + ":hoyos:" + metaname + ":builds"
        # Loda cache
        cache = await self.__get_cache(key)
        if cache:
            return Builds.parse_obj(cache)

        data = await self.__http.fetch_hoyos_by_username(
            profile_id, metaname, True)
        

        data = self.__format_json(data)

        self.LOGGER.debug("Parsing data...")

        # Store cache
        await self.__store_cache(key, data)

        return Builds.parse_obj(data)

    async def fetch_raw_data(self, uid: Union[str, int], *, info: bool = False) -> Dict[str, Any]:  # noqa
        """Fetches raw data for a user with the given UID. """

        # Loda cache
        cache = await self.__get_cache(uid)
        if cache:
            return cache

        data = await self.__http.fetch_user_by_uid(uid, info=info)
        data = self.__format_json(data)

        # Store cache
        await self.__store_cache(uid, data, cache=cache)

        return data

    async def sync_build(self, uid: Union[str, int], old_data: Dict[str, Any]) -> Dict[str, Any]:  # noqa
        """ Sync build data

        Parameters
        ----------
            uid: Union[:class:`str`,:class:`int`]
                The UID of the user to fetch data for.
            old_data: Dict[:class:`str`, Any]
                The build old data.

        Returns
        ------
            A dictionary containing the merged data.
        """

        new_data = await self.fetch_raw_data(uid)
        return await merge_raw_data(new_data, old_data)

    async def update_assets(self, path = None, lang: list = []) -> None:
        """ Updating Assets

        Args:
            path (str, optional): path to assets folder. Defaults to None.
            lang (list, optional): List of languages ​​that need to be updated, if [] then updates everything. Defaults to [].
        """
        if path is None:
            path = Assets._get_path_assets()
            path = os.path.dirname(path["data"])
        
        if not isinstance(lang, list):
            raise TypeError("The lang argument must be a list")
        
        await enka_update.dowload(path= path, lang = lang)
        await update_pfps(path = path)        
        self.assets.reload_assets()
        
    async def __format_hoyos(self, username: str, data: List[Any]) -> List[PlayerHoyos]:  # noqa
        return [PlayerHoyos.parse_obj({
            "builds": await self.fetch_builds(profile_id=username,
                                              metaname=data[key]["hash"]),
            **data[key]
        }) for key in data]

    def __format_json(self, data: Any):
        data = data["content"]
        return json.loads(data)

    async def __get_cache(
        self,
        cache_key: str,
    ):
        key = cache_key
        # Check config
        if Config.CACHE_ENABLED:
            self.LOGGER.warning(f"Getting data {key} from cache...")
            data = await Config.CACHE.get(key)

            if data is not None:
                self.LOGGER.debug("Parsing data...")
                return data

        return data

    async def __store_cache(self, key: str, data: Any, *, cache: Any = None):
        if Config.CACHE_ENABLED:
            self.LOGGER.debug(f"Caching data {key}...")
            if cache is None:
                await Config.CACHE.set(key, data)
            else:
                await Config.CACHE.set(key, await self.merge_raw_data(data,
                                                                      cache_data=cache))  # noqa

    # Concept by genshin.py python library
    fetch_user = fetch_user_by_uid
    fetch_profile = fetch_user_by_username
    merge_raw_data = merge_raw_data



from typing import List, Any, Dict

async def merge_raw_data(
    new_data: Dict[str, Any], 
    cache_data: Dict[str, Any]
) -> Dict[str, Any]:
    """
        Merge cached data into newly fetched data.

        Parameters
        ----------
            new_data: The newly fetched data as a dictionary.
            cache_data: The cached data as a dictionary.

        Returns
        -------
            A dictionary containing the merged data.
    """

    async def combine_lists(
        new_list: List[Dict[str, Any]], cache_list: List[Dict[str, Any]]
    ):
        new_ids = {item["avatarId"] for item in new_list}
        unique_cache_items = [
            item for item in cache_list if item["avatarId"] not in new_ids
        ]
        new_list.extend(unique_cache_items)

    if "showAvatarInfoList" in cache_data["playerInfo"]:
        new_data.setdefault("playerInfo", {}).setdefault(
            "showAvatarInfoList", [])
        await combine_lists(
            new_data["playerInfo"]["showAvatarInfoList"],
            cache_data["playerInfo"]["showAvatarInfoList"],
        )

    if "avatarInfoList" in cache_data:
        new_data.setdefault("avatarInfoList", [])
        await combine_lists(
            new_data["avatarInfoList"], cache_data["avatarInfoList"]
        )

    return new_data


"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz
Copyright (c) 2022-present M-307
Copyright (c) 2024-present DeviantUa

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import re
import logging
import sys
import os
import aiohttp
import json

from typing import Any, Dict, TYPE_CHECKING

from . import __version__

if TYPE_CHECKING:
    from aiohttp import ClientResponse

LOGGER = logging.getLogger(__name__)

# Base URL
BASE_URL = "https://enka.network/{PATH}"

# Request
CHUNK_SIZE = 5 * 2**20
RETRY_MAX = 10


def create_path(path: str) -> str:
    return BASE_URL.format(PATH=path)


def create_ui_path(filename: str) -> str:
    return create_path(f"ui/{filename}.png")


def validate_uid(uid: str) -> bool:
    """
        Validate UID
    """
    return len(uid) == 9 and uid.isdigit() and re.match(
        r"([1,2,5-9])\d{8}", uid)


def get_default_header():
    return {
        "User-Agent": get_user_agent(),
    }


def get_user_agent():
    # Get python version
    python_version = sys.version_info

    return "EnkaNetwork.py/{version} (Python {major}.{minor}.{micro})".format(
        version=__version__,
        major=python_version.major,
        minor=python_version.minor,
        micro=python_version.micro
    )


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __hash__(self):
        return 0

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()


async def to_data(response: ClientResponse) -> Dict[str, Any]:

    data = bytearray()
    data_to_read = True
    while data_to_read:
        red = 0
        while red < CHUNK_SIZE:
            chunk = await response.content.read(CHUNK_SIZE - red)

            if not chunk:
                data_to_read = False
                break

            data.extend(chunk)
            red += len(chunk)

    content = {
        "status": response.status,
        "content": data
    }
    return content


async def update_pfps(path):
    async with aiohttp.ClientSession() as session:
        async with session.get("https://raw.githubusercontent.com/EnkaNetwork/API-docs/master/store/pfps.json") as response:
            data = await response.read()
            
    data = json.loads(data)
    try:
        with open(os.path.join(path, "data/pfps.json"), 'w', encoding='utf-8') as file:
            json.dump(data, file, indent=4)
    except FileNotFoundError:
        with open(os.path.join(path, "pfps.json"), 'w', encoding='utf-8') as file:
            json.dump(data, file, indent=4)


from pydantic import BaseModel

from .character import CharacterInfo

from typing import Dict, List, Any, Optional, Union

class BuildInfo(BaseModel):
    id: int
    name: str = ''
    avatar_id: Union[str,bool] = ''
    avatar_data: CharacterInfo
    order: int = 0
    live: bool = False
    settings: Dict[str, Any] = {}
    public: bool = True
    
class Builds(BaseModel):
    raw: Dict[str, List[BuildInfo]] = {}

    def __init__(self, **data: Any) -> None:
        _data = {}
        for key in data:
            _build = []
            for build in data[key]:
                '''for equip in build["avatar_data"]["equipList"]:
                    print(equip)
                    equip["propValue"] = int(equip["propValue"])
                '''
                _build.append(build)
                #_build.append(build)

            _data[key] = _build

        super().__init__(**{"raw": _data})

    def get_build_info(self, avatar_id: str, build_id: str):
        buildList = self.raw.get(str(avatar_id), None)
        if buildList is None:
            return None
        
        for buildInfo in buildList:
            if str(buildInfo.id) == build_id:
                return buildInfo

        return None

    def get_character(self, avatar_id: str):
        return self.raw.get(str(avatar_id), None)

    def get_avatar_list(self):
        return [avatar_id for avatar_id in self.raw]


import logging

from pydantic import BaseModel, Field
from typing import List, Any, Union

from .utils import IconAsset

from ..assets import Assets
from ..enum import ElementType

LOGGER = logging.getLogger(__name__)

__all__ = (
    'ProfilePicture',
    'showAvatar',
    'Namecard',
    'PlayerInfo'
)

class ProfilePicture(BaseModel):
    """
        API Response data
    """
    id: Union[int, str] = Field(0, alias="avatarId")

    """
        Custom add data
    """
    icon: IconAsset = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.id = int(self.id)
        
        # Get character
        LOGGER.debug("=== Avatar ===")
        if "avatarId" in data:
            if "costumeId" in data:
                _data = Assets.character_costume(str(data["costumeId"]))
                icon = _data.images if _data is not None else _data
            else:
                icon = Assets.character_icon(str(data["avatarId"]))
        elif "id" in data:
            self.id = int(data["id"])
            _data = Assets.character_avatar(self.id)
            icon = _data.images if _data is not None else _data
            
            
        if not icon:
            return    
         
        self.icon = icon.icon


class showAvatar(BaseModel):
    """
        API Response data
    """
    id: Union[int, str] = Field(0, alias="avatarId")
    level: int = 1

    """
        Custom data
    """
    name: str = ""
    icon: IconAsset = None
    element: ElementType = ElementType.Unknown

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.id = int(self.id)
        
        # Get character
        LOGGER.debug("=== Character preview ===")

        # Find traveler
        character_preview = Assets.character(str(data["avatarId"]))

        if not character_preview:
            return

        self.element = character_preview.element

        if "costumeId" in data:
            _data = Assets.character_costume(str(data["costumeId"]))
            if _data:
                self.icon = _data.images.icon
        else:
            self.icon = character_preview.images.icon

        # Get name hash map
        _name = Assets.get_hash_map(str(character_preview.hash_id))

        if _name is None:
            return

        self.name = _name


class Namecard(BaseModel):
    id: int = 0
    name: str = ""
    icon: IconAsset = None
    banner: IconAsset = None
    navbar: IconAsset = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        if self.id > 0:
            LOGGER.debug("=== Namecard ===")
            # Get name card
            namecard = Assets.namecards(str(self.id))

            if not namecard:
                return

            self.icon = namecard.icon
            self.banner = namecard.banner
            self.navbar = namecard.navbar

            _name = Assets.get_hash_map(str(namecard.hash_id))
            if _name is None:
                return

            self.name = _name


class PlayerInfo(BaseModel):
    """
        API Response data
    """
    # Profile info
    achievement: int = Field(0, alias="finishAchievementNum")
    level: int = 0
    nickname: str = ""
    signature: str = ""
    world_level: int = Field(1, alias="worldLevel")
    avatar: ProfilePicture = Field(None, alias="profilePicture")
    # Avatars
    characters_preview: List[showAvatar] = Field([], alias="showAvatarInfoList")
    # Abyss floor
    abyss_floor: int = Field(0, alias="towerFloorIndex")
    abyss_room: int = Field(0, alias="towerLevelIndex")

    """
        Custom data
    """
    namecard: Namecard = Namecard()  # Profile name-card
    namecards: List[Namecard] = []  # List name-card preview in profile

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.namecard = Namecard(id=data["nameCardId"])
        self.namecards = [Namecard(id=namecard) for namecard in (data["showNameCardIdList"])] if "showNameCardIdList" in data else []  # noqa: E501



from pydantic import BaseModel, Field
from typing import List

from .players import PlayerInfo
from .character import CharacterInfo
from .profile import EnkaProfile, ProfileOwner
from .hoyos import PlayerHoyos

from ..utils import BASE_URL
from typing import Any

__all__ = ("EnkaNetworkResponse",)

class EnkaNetworkInfo(BaseModel):
    """ Enka.Network response data

    Attributes
    ------------
    uid: :class:`str`
        UID Player
    url: :class:`str`
        URL to Enka.Network profile
    path: :class:`str`
        Path to enter Enka.network profile
    """
    uid: str = ""
    url: str = ""
    path: str = ""

class EnkaNetworkResponse(BaseModel):
    """ Enka.Network response data

    Attributes
    ------------
    player: :class:`PlayerInfo`
        Player info data. Please refers in `PlayerInfo` class (key: "playerInfo")
    characters: List[:class:`CharacterInfo`]
        List character. Please refers in `CharacterInfo` class (key: "avatarInfoList")
    profile: :class:`EnkaNetworkInfo`
        profile enka.network. Please refers in `EnkaNetworkInfo` class
    owner: :class:`ProfileOwner`
        Owner UID data. **subscriptions in Enka.Network**. Please refers in `ProfileOwner` class 
    ttl: :class:`int`
        Cache timeout
    uid: :class:`int`
        UID Player
    """
    player: PlayerInfo = Field(None, alias="playerInfo")
    characters: List[CharacterInfo] = Field(None, alias="avatarInfoList")
    profile: EnkaNetworkInfo = EnkaNetworkInfo()
    owner: ProfileOwner = None
    ttl: int = 0
    uid: int = 0

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.profile.path = f"/u/{data.get('uid')}"
        self.profile.url = BASE_URL.format(PATH=self.profile.path[1:])
        self.profile.uid = data.get("uid") or ""

class EnkaNetworkProfileResponse(BaseModel):
    username: str
    profile: EnkaProfile
    hoyos: List[PlayerHoyos] = []


from pydantic import BaseModel

from ..enum import ProfileRank
from .build import Builds

class EnkaProfile(BaseModel):
    bio: str = ''
    level: ProfileRank = ProfileRank.TIER_1
    signup_state: int = 0
    image_url: str = ''

class ProfileOwner(BaseModel):
    hash: str
    username: str
    profile: EnkaProfile
    builds: Builds = None


import math

from pydantic import BaseModel

from typing import Any

__all__ = (
    'Stats',
    'StatsPercentage',
    'CharacterStats'
)

class Stats(BaseModel):
    id: int = 0
    value: float = 0

    def to_rounded(self) -> float:
        return math.ceil(self.value)


class StatsPercentage(BaseModel):
    id: int = 0
    value: float = 0.0

    def to_percentage(self) -> float:
        return round((round(self.value, 3) * 100), 1)

    def to_percentage_symbol(self) -> str:
        return f"{self.to_percentage()}%"


class CharacterStats(BaseModel):
    BASE_HP: Stats = Stats(id=1, value=0)
    FIGHT_PROP_HP: Stats = Stats(id=2, value=0)
    FIGHT_PROP_HP_PERCENT: StatsPercentage = StatsPercentage(id=3, value=0)
    FIGHT_PROP_BASE_ATTACK: Stats = Stats(id=4, value=0)
    FIGHT_PROP_ATTACK: Stats = Stats(id=5, value=0)
    FIGHT_PROP_ATTACK_PERCENT: StatsPercentage = StatsPercentage(id=6, value=0)
    FIGHT_PROP_BASE_DEFENSE: Stats = Stats(id=7, value=0)
    FIGHT_PROP_DEFENSE: Stats = Stats(id=8, value=0)
    FIGHT_PROP_DEFENSE_PERCENT: StatsPercentage = StatsPercentage(id=9, value=0)
    FIGHT_PROP_BASE_SPEED: Stats = Stats(id=10, value=0)
    FIGHT_PROP_SPEED_PERCENT: StatsPercentage = StatsPercentage(id=11, value=0)
    FIGHT_PROP_HP_MP_PERCENT: StatsPercentage = StatsPercentage(id=12, value=0)
    FIGHT_PROP_ATTACK_MP_PERCENT: Stats = Stats(id=13, value=0)
    FIGHT_PROP_CRITICAL: StatsPercentage = StatsPercentage(id=20, value=0)
    FIGHT_PROP_ANTI_CRITICAL: StatsPercentage = StatsPercentage(id=21, value=0)
    FIGHT_PROP_CRITICAL_HURT: StatsPercentage = StatsPercentage(id=22, value=0)
    FIGHT_PROP_CHARGE_EFFICIENCY: StatsPercentage = StatsPercentage(id=23, value=0)
    FIGHT_PROP_ADD_HURT: StatsPercentage = StatsPercentage(id=24, value=0)
    FIGHT_PROP_SUB_HURT: StatsPercentage = StatsPercentage(id=25, value=0)
    FIGHT_PROP_HEAL_ADD: StatsPercentage = StatsPercentage(id=26, value=0)
    FIGHT_PROP_HEALED_ADD: StatsPercentage = StatsPercentage(id=27, value=0)
    FIGHT_PROP_ELEMENT_MASTERY: Stats = Stats(id=28, value=0)
    FIGHT_PROP_PHYSICAL_SUB_HURT: StatsPercentage = StatsPercentage(id=29, value=0)
    FIGHT_PROP_PHYSICAL_ADD_HURT: StatsPercentage = StatsPercentage(id=30, value=0)
    FIGHT_PROP_DEFENCE_IGNORE_RATIO: Stats = Stats(id=31, value=0)
    FIGHT_PROP_DEFENCE_IGNORE_DELTA: Stats = Stats(id=32, value=0)
    FIGHT_PROP_FIRE_ADD_HURT: StatsPercentage = StatsPercentage(id=40, value=0)
    FIGHT_PROP_ELEC_ADD_HURT: StatsPercentage = StatsPercentage(id=41, value=0)
    FIGHT_PROP_WATER_ADD_HURT: StatsPercentage = StatsPercentage(id=42, value=0)
    FIGHT_PROP_GRASS_ADD_HURT: StatsPercentage = StatsPercentage(id=43, value=0)
    FIGHT_PROP_WIND_ADD_HURT: StatsPercentage = StatsPercentage(id=44, value=0)
    FIGHT_PROP_ROCK_ADD_HURT: StatsPercentage = StatsPercentage(id=45, value=0)
    FIGHT_PROP_ICE_ADD_HURT: StatsPercentage = StatsPercentage(id=46, value=0)
    FIGHT_PROP_HIT_HEAD_ADD_HURT: StatsPercentage = StatsPercentage(id=47, value=0)
    FIGHT_PROP_FIRE_SUB_HURT: StatsPercentage = StatsPercentage(id=50, value=0)
    FIGHT_PROP_ELEC_SUB_HURT: StatsPercentage = StatsPercentage(id=51, value=0)
    FIGHT_PROP_WATER_SUB_HURT: StatsPercentage = StatsPercentage(id=52, value=0)
    FIGHT_PROP_GRASS_SUB_HURT: StatsPercentage = StatsPercentage(id=53, value=0)
    FIGHT_PROP_WIND_SUB_HURT: StatsPercentage = StatsPercentage(id=54, value=0)
    FIGHT_PROP_ROCK_SUB_HURT: StatsPercentage = StatsPercentage(id=55, value=0)
    FIGHT_PROP_ICE_SUB_HURT: StatsPercentage = StatsPercentage(id=56, value=0)
    FIGHT_PROP_EFFECT_HIT: Stats = Stats(id=60, value=0)
    FIGHT_PROP_EFFECT_RESIST: Stats = Stats(id=61, value=0)
    FIGHT_PROP_FREEZE_RESIST: Stats = Stats(id=62, value=0)
    FIGHT_PROP_TORPOR_RESIST: Stats = Stats(id=63, value=0)
    FIGHT_PROP_DIZZY_RESIST: Stats = Stats(id=64, value=0)
    FIGHT_PROP_FREEZE_SHORTEN: Stats = Stats(id=65, value=0)
    FIGHT_PROP_TORPOR_SHORTEN: Stats = Stats(id=66, value=0)
    FIGHT_PROP_DIZZY_SHORTEN: Stats = Stats(id=67, value=0)
    FIGHT_PROP_MAX_FIRE_ENERGY: Stats = Stats(id=70, value=0)
    FIGHT_PROP_MAX_ELEC_ENERGY: Stats = Stats(id=71, value=0)
    FIGHT_PROP_MAX_WATER_ENERGY: Stats = Stats(id=72, value=0)
    FIGHT_PROP_MAX_GRASS_ENERGY: Stats = Stats(id=73, value=0)
    FIGHT_PROP_MAX_WIND_ENERGY: Stats = Stats(id=74, value=0)
    FIGHT_PROP_MAX_ICE_ENERGY: Stats = Stats(id=75, value=0)
    FIGHT_PROP_MAX_ROCK_ENERGY: Stats = Stats(id=76, value=0)
    FIGHT_PROP_SKILL_CD_MINUS_RATIO: Stats = Stats(id=80, value=0)
    FIGHT_PROP_SHIELD_COST_MINUS_RATIO: Stats = Stats(id=81, value=0)

    FIGHT_PROP_CUR_FIRE_ENERGY: Stats = Stats(id=1000, value=0)
    FIGHT_PROP_CUR_ELEC_ENERGY: Stats = Stats(id=1001, value=0)
    FIGHT_PROP_CUR_WATER_ENERGY: Stats = Stats(id=1002, value=0)
    FIGHT_PROP_CUR_GRASS_ENERGY: Stats = Stats(id=1003, value=0)
    FIGHT_PROP_CUR_WIND_ENERGY: Stats = Stats(id=1004, value=0)
    FIGHT_PROP_CUR_ICE_ENERGY: Stats = Stats(id=1005, value=0)
    FIGHT_PROP_CUR_ROCK_ENERGY: Stats = Stats(id=1006, value=0)
    FIGHT_PROP_CUR_HP: Stats = Stats(id=1010, value=0)

    FIGHT_PROP_MAX_HP: Stats = Stats(id=2000, value=0)
    FIGHT_PROP_CUR_ATTACK: Stats = Stats(id=2001, value=0)
    FIGHT_PROP_CUR_DEFENSE: Stats = Stats(id=2002, value=0)
    FIGHT_PROP_CUR_SPEED: Stats = Stats(id=2003, value=0)

    FIGHT_PROP_NONEXTRA_ATTACK: Stats = Stats(id=3000, value=0)
    FIGHT_PROP_NONEXTRA_DEFENSE: Stats = Stats(id=3001, value=0)
    FIGHT_PROP_NONEXTRA_CRITICAL: StatsPercentage = StatsPercentage(id=3002, value=0)
    FIGHT_PROP_CUR_SPEED: Stats = Stats(id=3003, value=0)
    FIGHT_PROP_NONEXTRA_CRITICAL_HURT: StatsPercentage = StatsPercentage(id=3004, value=0)
    FIGHT_PROP_NONEXTRA_CHARGE_EFFICIENCY: StatsPercentage = StatsPercentage(id=3005, value=0)
    FIGHT_PROP_NONEXTRA_ELEMENT_MASTERY: Stats = Stats(id=3006, value=0)
    FIGHT_PROP_NONEXTRA_PHYSICAL_SUB_HURT: StatsPercentage = StatsPercentage(id=3007, value=0)
    FIGHT_PROP_NONEXTRA_FIRE_ADD_HURT: StatsPercentage = StatsPercentage(id=3008, value=0)
    FIGHT_PROP_NONEXTRA_ELEC_ADD_HURT: StatsPercentage = StatsPercentage(id=3009, value=0)
    FIGHT_PROP_NONEXTRA_WATER_ADD_HURT: StatsPercentage = StatsPercentage(id=3010, value=0)
    FIGHT_PROP_NONEXTRA_GRASS_ADD_HURT: StatsPercentage = StatsPercentage(id=3011, value=0)
    FIGHT_PROP_NONEXTRA_WIND_ADD_HURT: StatsPercentage = StatsPercentage(id=3012, value=0)
    FIGHT_PROP_NONEXTRA_ROCK_ADD_HURT: StatsPercentage = StatsPercentage(id=3013, value=0)
    FIGHT_PROP_NONEXTRA_ICE_ADD_HURT: StatsPercentage = StatsPercentage(id=3014, value=0)
    FIGHT_PROP_NONEXTRA_FIRE_SUB_HURT: StatsPercentage = StatsPercentage(id=3015, value=0)
    FIGHT_PROP_NONEXTRA_ELEC_SUB_HURT: StatsPercentage = StatsPercentage(id=3016, value=0)
    FIGHT_PROP_NONEXTRA_WATER_SUB_HURT: StatsPercentage = StatsPercentage(id=3017, value=0)
    FIGHT_PROP_NONEXTRA_GRASS_SUB_HURT: StatsPercentage = StatsPercentage(id=3018, value=0)
    FIGHT_PROP_NONEXTRA_WIND_SUB_HURT: StatsPercentage = StatsPercentage(id=3019, value=0)
    FIGHT_PROP_NONEXTRA_ROCK_SUB_HURT: StatsPercentage = StatsPercentage(id=3020, value=0)
    FIGHT_PROP_NONEXTRA_ICE_SUB_HURT: StatsPercentage = StatsPercentage(id=3021, value=0)
    FIGHT_PROP_NONEXTRA_SKILL_CD_MINUS_RATIO: Stats = Stats(id=3022, value=0)
    FIGHT_PROP_NONEXTRA_SHIELD_COST_MINUS_RATIO: Stats = Stats(id=3023, value=0)
    FIGHT_PROP_NONEXTRA_PHYSICAL_ADD_HURT: StatsPercentage = StatsPercentage(id=3024, value=0)

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        _stats = self.__dict__
        for key in _stats:
            _stats[key].value = data[str(_stats[key].id)] if str(_stats[key].id) in data else 0



import logging

from pydantic import BaseModel, Field
from typing import List, Any, Dict, Union

from .equipments import Equipments
from .stats import CharacterStats
from .assets import (
    CharacterIconAsset,
)
from .utils import IconAsset

from ..assets import Assets
from ..enum import ElementType

LOGGER = logging.getLogger(__name__)

__all__ = (
    'CharacterSkill',
    'CharacterConstellations',
    'CharacterInfo'
)

class CharacterSkill(BaseModel):
    id: int = 0
    name: str = ""
    icon: IconAsset = None
    is_boosted: bool = False
    level: int = 0


class CharacterConstellations(BaseModel):
    id: int = 0
    name: str = ""
    icon: IconAsset = None
    unlocked: bool = False  # If character has this constellation.


class CharacterInfo(BaseModel):
    """
        API Response data
    """
    id: Union[int, str] = Field(0, alias="avatarId")
    equipments: List[Equipments] = Field([], alias="equipList")
    stats: CharacterStats = Field({}, alias="fightPropMap")
    skill_data: List[int] = Field([], alias="inherentProudSkillList")
    skill_id: int = Field(0, alias="skillDepotId")

    """
        Custom data
    """
    name: str = ""  # Get from name hash map
    friendship_level: int = 1
    element: ElementType = ElementType.Unknown
    rarity: int = 0
    image: CharacterIconAsset = None
    skills: List[CharacterSkill] = []
    constellations: List[CharacterConstellations] = []

    # Prop Maps
    xp: int = 0  # AKA. propMap 1001
    ascension: int = 0  # AKA. propMap 4001
    level: int = 0  # AKA. propMap 1002

    # Other
    max_level: int = 20
    constellations_unlocked: int = 0  # Constellation is unlocked count

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        self.id = int(self.id)
        # Friendship level
        self.friendship_level = data["fetterInfo"]["expLevel"]

        # Get prop map
        self.xp = int(data["propMap"]["1001"].get("ival", 0)) if "1001" in data["propMap"] else 0
        self.ascension = int(data["propMap"]["1002"].get("ival", 0)) if "1002" in data["propMap"] else 0
        self.level = int(data["propMap"]["4001"].get("ival", 0)) if "4001" in data["propMap"] else 0

        # Constellation unlocked count
        self.constellations_unlocked = len(data["talentIdList"]) if "talentIdList" in data else 0

        # Get max character level
        self.max_level = (self.ascension * 10) + (10 if self.ascension > 0 else 0) + 20

        # Get character
        LOGGER.debug("=== Character Data ===")
        avatarId = str(data["avatarId"])
        avatarId += f"-{data['skillDepotId']}" if data["avatarId"] in [10000005, 10000007] else ""
        character = Assets.character(avatarId)

        # Check if character is founded
        if not character:
            return

        # Load icon
        if "costumeId" in data:
            _data = Assets.character_costume(str(data["costumeId"]))
            if _data:
                self.image = _data.images
            else:
                self.image = character.images
        else:
            self.image = character.images

        # Get element
        self.element = ElementType(character.element)

        # Check calculate rarity
        self.rarity = character.rarity

        # Load constellation
        LOGGER.debug("=== Constellation ===")
        for constellation in character.constellations:
            _constellation = Assets.constellations(constellation)
            if not _constellation:
                continue

            # Get name hash map
            _name = Assets.get_hash_map(str(_constellation.hash_id))

            if _name is None:
                continue

            self.constellations.append(CharacterConstellations(
                id=int(constellation),
                name=_name,
                icon=_constellation.icon,
                unlocked=int(
                    constellation) in data["talentIdList"] if "talentIdList" in data else False
            ))

        # Load skills
        LOGGER.debug("=== Skills ===")
        for skill in character.skills:
            _lvl = data["skillLevelMap"].get(str(skill), 0)
            _is_boosted = False
            _skill = Assets.skills(skill)

            if not _skill:
                continue

            # Get name hash map
            _name = Assets.get_hash_map(_skill.hash_id)

            if _name is None:
                continue

            if "proudSkillExtraLevelMap" in data:
                boost_level = data["proudSkillExtraLevelMap"].get(str(_skill.pround_map), None)
                if not boost_level is None:
                    _is_boosted = True 
                    _lvl += boost_level

            self.skills.append(CharacterSkill(
                id=skill,
                name=_name,
                icon=_skill.icon,
                is_boosted=_is_boosted,
                level=_lvl
            ))

        LOGGER.debug("=== Character Name ===")
        _name = Assets.get_hash_map(str(character.hash_id))

        if _name is None:
            return

        # Get name from hash map
        self.name = _name

    class Config:
        use_enum_values = True



import json
import os
import logging

from .enum import Language
from .model import assets, utils
from .utils import create_ui_path

from typing import Dict, List, TextIO, Optional, Union, Callable

PATH = os.path.dirname(os.path.abspath(__file__))

LOGGER = logging.getLogger(__name__)

__all__ = ('Assets',)


def check_assets(f: Callable):
    def decorator(*args):
        if not args[0].DATA or \
           not args[0].HASH_MAP:
           args[0]._set_language("EN")
           args[0].reload_assets()
        
        return f(*args)

    return decorator


class Assets:
    DATA: Dict[str, dict] = {}
    HASH_MAP: Dict[str, dict] = {}
    LANGS: Language = Language.EN

    def __init__(self, lang: Union[str, Language] = Language.EN) -> None:
        # Set language
        self._set_language(lang)
        self.reload_assets()

    @classmethod
    def reload_assets(cls) -> None:
        # Load assets
        cls.__load_assets_lang()
        cls.__load_assets_data()

    @property
    def CHARACTERS_IDS(self) -> List[str]:
        return [x for x in self.DATA["characters"]]

    @property
    def COSTUMES_IDS(self) -> List[str]:
        return [x for x in self.DATA["costumes"]]

    @property
    def NAMECARD_IDS(self) -> List[str]:
        return [x for x in self.DATA["namecards"]]

    @classmethod
    @check_assets
    def character(cls, id: Union[int, str]) -> Optional[assets.CharacterAsset]:
        LOGGER.debug(f"Getting character assets with id: {id}")

        data = cls.DATA["characters"].get(str(id))
        if not data:
            LOGGER.error(f"Character not found with id: {id}")
            return

        return assets.CharacterAsset.parse_obj({
            "id": id if str(id).isdigit() else id.split("-")[0],
            "skill_id": str(id).split("-")[1] if not str(id).isdigit() else 0,
            "images": cls.create_character_icon(data["sideIconName"]),
            **data
        })

    @classmethod
    def character_avatar(cls, id: int) -> Optional[assets.CharacterCostume]:
        LOGGER.debug(f"Getting Avatar assets with id: {id}")
        data = cls.DATA["pfps"].get(str(id))
        if not data:
            LOGGER.error(f"Avatar not found with id: {id}")
            return        
        return assets.CharacterAvatar.parse_obj({
            "id": id,
            "images": cls.create_character_icon(data["iconPath"])
        })
    
    @classmethod
    def character_costume(cls, id: int) -> Optional[assets.CharacterCostume]:
        LOGGER.debug(f"Getting costume assets with id: {id}")
        data = cls.DATA["costumes"].get(str(id))
        if not data:
            LOGGER.error(f"Costume not found with id: {id}")
            return

        return assets.CharacterCostume.parse_obj({
            "id": id,
            "images": cls.create_chractar_costume_icon(data["sideIconName"])
        })

    @classmethod
    def constellations(cls, id: int) -> Optional[assets.CharacterConstellationsAsset]:
        LOGGER.debug(f"Getting character constellations assets with id: {id}")
        data = cls.DATA["constellations"].get(str(id))
        if not data:
            LOGGER.error(f"Character constellations not found with id: {id}")
            return

        return assets.CharacterConstellationsAsset.parse_obj({
            "id": id,
            **data,
            "icon": utils.IconAsset(filename=data["icon"])
        })

    @classmethod
    def skills(cls, id: int) -> Optional[assets.CharacterSkillAsset]:
        LOGGER.debug(f"Getting character skills assets with id: {id}")
        data = cls.DATA["skills"].get(str(id))

        if not data:
            LOGGER.error(f"Character skills not found with id: {id}")
            return


        pround = data.get("proudSkillGroupId", 0)
        return assets.CharacterSkillAsset.parse_obj({
            "id": id,
            **data,
            "pround_map": pround if not pround is None and pround != "" else 0,
            "icon": utils.IconAsset(filename=data["skillIcon"])
        })

    @classmethod
    def namecards(cls, id: int) -> Optional[assets.NamecardAsset]:
        LOGGER.debug(f"Getting namecards assets with id: {id}")
        data = cls.DATA["namecards"].get(str(id))
        if not data:
            LOGGER.error(f"Namecards not found with id: {id}")
            return

        return assets.NamecardAsset.parse_obj({
            "id": id,
            **data,
            "icon": utils.IconAsset(filename=data["icon"]),
            "banner": utils.IconAsset(filename=data["picPath"][1]),
            "navbar": utils.IconAsset(filename=data["picPath"][0]),
        })
    
    @classmethod
    def artifact_props(cls, id: int):
        LOGGER.debug(f"Getting artifact props assets with id: {id}")
        data = cls.DATA["artifact_props"].get(str(id))
        if not data:
            LOGGER.error(f"Artifact props not found with id: {id}")
            return
        
        return assets.AritfactProps.parse_obj({
            "id": id,
            **data
        })

    @classmethod
    def get_hash_map(cls, hash_id: str) -> Optional[str]:
        LOGGER.debug(f"Getting nameTextMapHash {hash_id} with language: {cls.LANGS}")  # noqa: E501
        for key in cls.HASH_MAP:
            if str(hash_id) in cls.HASH_MAP[key]:
                val = cls.HASH_MAP[key][str(hash_id)][cls.LANGS]
                LOGGER.debug(f"Got nameTextMapHash {hash_id} with language: {key} (Value '{val}')")  # noqa: E501
                return val

        LOGGER.error(f"nameTextMapHash {hash_id} not found with language: {cls.LANGS}")  # noqa: E501
        return

    @classmethod
    def character_icon(cls, id: int) -> Optional[assets.CharacterIconAsset]:
        data = cls.character(id)
        if not data:
            return

        return data.images
    
    @classmethod
    def character_icon_pfps(cls, id: int) -> Optional[assets.CharacterIconAsset]:
        data = cls.character(id)
        if not data:
            return

        return data.images

    @staticmethod
    def create_character_icon(path: str) -> assets.CharacterIconAsset:
        return assets.CharacterIconAsset(
            icon=utils.IconAsset(filename=path.replace("_Side", "")),
            side=utils.IconAsset(filename=path),
            banner=utils.IconAsset(filename=path.replace("AvatarIcon_Side", "Gacha_AvatarImg")),  # noqa: E501
            card=utils.IconAsset(filename=path.replace("_Side", "") + "_Card")
        )
        
         

    @classmethod
    def create_chractar_costume_icon(cls, path: str) -> assets.CharacterIconAsset:  # noqa: E501
        _data = cls.create_character_icon(path)
        _data.banner = utils.IconAsset(filename=_data.banner.filename.replace("Gacha_AvatarImg", "Costume"))  # noqa: E501
        return _data

    @staticmethod
    def create_icon_path(path: str) -> str:
        return create_ui_path(path)

    @classmethod
    def _set_language(cls, lang: Language) -> None:
        # Check language
        if lang is None or not lang.split("-")[0].lower() in list(Language):
            raise ValueError("Language not supported. Please check your language.")  # noqa: E501

        LOGGER.debug(f"Set language to {lang}.")
        cls.LANGS = lang.upper()

    @classmethod
    def _get_path_assets(cls) -> Dict[str, str]:
        return {
            "data": os.path.join(PATH, "assets", "data"),
            "langs": os.path.join(PATH, "assets", "langs")
        }

    @classmethod
    def __load_assets_lang(cls) -> None:
        _PATH = cls._get_path_assets()["langs"]
        FILE_LANG = os.listdir(_PATH)
        for FILENAME in FILE_LANG:
            LOGGER.debug(f"Loading language file {FILENAME}...")
            cls.HASH_MAP[FILENAME.split(".")[0]] = json.load(cls.__load(os.path.join(_PATH, FILENAME)))  # noqa: E501

    @classmethod
    def __load_assets_data(cls) -> None:
        _PATH = cls._get_path_assets()["data"]
        FILE_DATA = os.listdir(_PATH)
        for FILENAME in FILE_DATA:
            LOGGER.debug(f"Loading data file {FILENAME}...")

            cls.DATA[FILENAME.split(".")[0]] = json.load(cls.__load(os.path.join(_PATH, FILENAME)))  # noqa: E501

    @staticmethod
    def __load(path: str) -> TextIO:
        return open(path, "r", encoding="utf-8")



from enum import Enum

__all__ = (
    'Language',
    'EquipmentsType',
    'DigitType',
    'ElementType',
    'EquipType'
)

class Language(str, Enum):
    EN = "en"
    RU = "ru"
    VI = "vi"
    TH = "th"
    PT = "pt"
    KR = "kr"
    JP = "jp"
    ID = "id"
    FR = "fr"
    ES = "es"
    DE = "de"
    IT = "it"
    TR = "tr"

    """
        zh-CN: CHT
        zh-TW: CHS
    """
    TW = "cht"
    CN = "chs"
    CHT = "cht"
    CHS = "chs"


class EquipmentsType(int, Enum):
    UNKNOWN = -1
    ARTIFACT = 0
    WEAPON = 1


class DigitType(int, Enum):
    NUMBER = 0
    PERCENT = 1


class ElementType(str, Enum):
    Unknown = "Unknown"
    Cryo = "Ice"
    Hydro = "Water"
    Anemo = "Wind"
    Pyro = "Fire"
    Geo = "Rock"
    Electro = "Electric"
    Dendro = "Grass"

class EquipType(str, Enum):
    Unknown = "UNKNOWN"
    Flower = "EQUIP_BRACER"
    Feather = "EQUIP_NECKLACE"
    Sands = "EQUIP_SHOES"
    Goblet = "EQUIP_RING"
    Circlet = "EQUIP_DRESS"

class ProfileRank(int, Enum):
    TIER_LEGACY = -1
    TIER_NONE = 0
    TIER_1 = 1
    TIER_2 = 2
    TIER_3 = 3



import json

from typing import Any, Dict, Optional
from cachetools import TTLCache

__all__ = ('StaticCache','Cache')

class Cache:
    async def get(self, key: str) -> Optional[Dict[str, Any]]:
        pass

    async def set(self, key: str, value: Dict[str, Any]) -> None:
        pass

class StaticCache:
    def __init__(self, maxsize: int, ttl: int) -> None:
        self.cache = TTLCache(maxsize, ttl)

    async def get(self, key) -> Dict[str, Any]:
        data = self.cache.get(key)
        return json.loads(data) if data is not None else data

    async def set(self, key, value) -> None:
        self.cache[key] = json.dumps(value)



import setuptools
import re

with open('enkanetwork/__init__.py') as f:
	"""
		Get version from utils.py
		Ref: https://github.com/Rapptz/discord.py/blob/52f3a3496bea13fefc08b38f9ed01641e565d0eb/setup.py#L9
	"""
	version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)

setuptools.setup(
	name="enkanetworkv2.py",
	version=version,
	author="DeviantUa",
	author_email="deviantapi@gmail.com",
	description="Library for fetching JSON data from site https://enka.network/",
	long_description=open("README.md", "r", encoding="utf-8").read(),
	long_description_content_type="text/markdown",
	url="https://github.com/DEViantUA/EnkaNetworkV2.py",
	keywords = ['enkanetwork.py', 'enkanetwork', 'enka.network', 'genshinapi', 'enkanetworkV2', 'enkanetworkV2.py'],
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	install_requires=[
		"pydantic",
		"aiohttp",
		"cachetools",
  		"enkacard"
	],
	python_requires=">=3.6",
	include_package_data=True
)




from typing import (
    ClassVar
)
from .utils import get_user_agent
from .cache import Cache, StaticCache

class Config:
    # HTTP Config
    ENKA_PROTOCOL: ClassVar[str] = "https"
    ENKA_URL: ClassVar[str] = "enka.network"
    # Assets
    ASSETS_PROTOCOL: ClassVar[str] = "https"
    ASSETS_URL: ClassVar[str] = "raw.githubusercontent.com"
    # Header Config
    USER_AGENT: ClassVar[str] = get_user_agent()
    # Client config
    CACHE_ENABLED: ClassVar[bool] = True
    CACHE: ClassVar[Cache] = StaticCache(1024, 60 * 3)

    @classmethod
    def init_cache(
        cls, 
        cache: Cache,
        enabled: bool = True
    ):
        cls.CACHE = cache
        cls.CACHE_ENABLED = enabled
    
    @classmethod
    def init_user_agent(cls, agent: str):
        cls.USER_AGENT = agent